package work.xuye.source.handler;

import com.google.gson.Gson;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cloud.stream.function.StreamBridge;
import org.springframework.expression.EvaluationContext;
import org.springframework.expression.Expression;
import org.springframework.expression.spel.standard.SpelExpressionParser;
import org.springframework.expression.spel.support.StandardEvaluationContext;
import org.springframework.messaging.Message;
import org.springframework.messaging.support.MessageBuilder;
import org.springframework.stereotype.Component;
import org.springframework.util.DigestUtils;
import org.springframework.util.ObjectUtils;
import work.xuye.common.constant.*;
import work.xuye.common.db.entity.Task;
import work.xuye.common.db.entity.vo.NextTaskFilter;
import work.xuye.common.db.entity.vo.TaskMiniVO;
import work.xuye.common.db.service.TaskManager;
import work.xuye.common.dto.HttpRequestParams;
import work.xuye.common.dto.HttpRes;
import work.xuye.common.enums.ProcessMode;
import work.xuye.common.enums.RequestMode;
import work.xuye.common.enums.ResourceStatus;
import work.xuye.common.enums.SignatureEnumGroup;
import work.xuye.common.service.UrlMD5Service;
import work.xuye.common.utils.DebugUtil;
import work.xuye.common.utils.ID;
import work.xuye.common.utils.JsonPathUtil;
import work.xuye.common.utils.signature.AbstractSignatureService;
import work.xuye.common.utils.signature.SignatureFactory;
import work.xuye.source.request.RequestClient;
import work.xuye.source.spel.CustomRequest;

import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.util.*;
import java.util.regex.Matcher;
import java.util.stream.Stream;

/**
 * @author xuye
 * @since 2023/2/19 16:37
 **/
@Slf4j
@Component
@RequiredArgsConstructor
public class SourceHandler {

    public final CustomRequest cr;
    private final SpelExpressionParser parser;
    private final RequestClient requestClient;
    private final UrlMD5Service urlMD5Service;
    private final StreamBridge streamBridge;
    private final TaskManager taskManager;
    public final Gson gson;

    public List<Message<HttpRes>> handle(Message<TaskMiniVO> message) {
        TaskMiniVO taskMiniVO = message.getPayload();
        if (DebugUtil.isSkip(taskMiniVO)) {
            return null;
        }
        Task task = taskManager.getTaskInfoByTaskId(taskMiniVO.getId()).getTask();
        if (DebugUtil.isSkip(task)) {
            return null;
        }
        HttpRequestParams req = task.getRequestParams();
        if (!ObjectUtils.isEmpty(taskMiniVO.getArgs())) {
            req.getArgs().putAll(taskMiniVO.getArgs());
        }

        Set<String> otherCacheKeySet = new HashSet<>();
        Set parentOtherCacheKeySet = message.getHeaders().get(MessageConstants.OTHER_CACHE_KEY_SET, Set.class);
        if (parentOtherCacheKeySet != null) {
            otherCacheKeySet.addAll(parentOtherCacheKeySet);
        }


        String taskName = task.getName();
        HttpRes res = null;
        List<HttpRes> result = null;
        try {
            if (RequestMode.normal.equals(req.getMode())) {
                res = this.request(req);
                result = this.handleRes(task, res);
            } else if (RequestMode.SpEL.equals(req.getMode())) {

                StandardEvaluationContext context = new StandardEvaluationContext();
                Expression expression = parser.parseExpression(task.getRequestParams().getSpELConfig().getExpression());
                List<HttpRes> value = expression.getValue(context, this, List.class);
                result = this.handleRes(task, value);
            }
        } catch (Exception e) {
            urlMD5Service.removeUrlCache(message);
        }
        if (ObjectUtils.isEmpty(result)) {
            return null;
        }
        if (ProcessMode.WATCH.equals(task.getProcessMode())) {
            for (HttpRes r : result) {
                Message<String> watchMessage = MessageBuilder
                        .withPayload(r.getBody())
                        .setHeader(MessageConstants.CURRENT_STAGE, StageConstants.SOURCE_WATCH)
                        .setHeader(MessageConstants.TASK_TRACE_ID, message.getHeaders().get(MessageConstants.TASK_TRACE_ID))
                        .setHeader(MessageConstants.SOURCE_TRACE_ID, ID.generate())
                        .setHeader(MessageConstants.TASK_ID, task.getId())
                        .setHeader(MessageConstants.TASK_NAME, task.getName())
                        .setHeader(SnapshotConstants.URL, req.getUrl())
                        .setHeader(SnapshotConstants.RESOURCE_STATUS, r.getResourceStatus())
                        .setHeader(SnapshotConstants.STATUS, r.getStatus())
                        .setHeader(MessageConstants.OTHER_CACHE_KEY_SET, otherCacheKeySet)
                        .build();

                boolean send = streamBridge.send(taskName + BindingConstants.SNAPSHOT_OUT_POSTFIX, watchMessage);
                if (!send) {
                    throw new RuntimeException("send message failed");
                }
            }
            return null;
        }

        if (Objects.nonNull(task.getNextFilter())) {

            Map<String, Message> nextNodeMessage = buildNextNodeMessage(result, task, message);

            if (!ObjectUtils.isEmpty(nextNodeMessage)) {
                nextNodeMessage.forEach((k, v) -> {
                    Set set = v.getHeaders().get(MessageConstants.OTHER_CACHE_KEY_SET, Set.class);
                    if (!ObjectUtils.isEmpty(set) && !ObjectUtils.isEmpty(otherCacheKeySet)) {
                        set.addAll(otherCacheKeySet);
                        v.getHeaders().put(MessageConstants.OTHER_CACHE_KEY_SET, set);
                    }
                    boolean send = streamBridge.send(k, v);
                    if (!send) {
                        throw new RuntimeException("send message failed");
                    }
                });
            }
            return null;
        }

        ArrayList<Message<HttpRes>> messageList = new ArrayList<>();
        result.forEach(
                item -> {
                    messageList.add(MessageBuilder
                            .withPayload(item)
                            .setHeader(MessageConstants.CURRENT_STAGE, StageConstants.SOURCE)
                            .setHeader(MessageConstants.TASK_TRACE_ID, message.getHeaders().get(MessageConstants.TASK_TRACE_ID))
                            .setHeader(MessageConstants.SOURCE_TRACE_ID, ID.generate())
                            .setHeader(MessageConstants.TASK_ID, task.getId())
                            .setHeader(MessageConstants.TASK_NAME, task.getName())
                            .setHeader(MessageConstants.SEED_URL, item.getRequestParams().getUrl())
                            .setHeader(MessageConstants.OTHER_CACHE_KEY_SET, otherCacheKeySet)
                            .build());
                }
        );
        return messageList;


    }

    private Map<String, Message> buildNextNodeMessage(List<HttpRes> result, Task task, Message<TaskMiniVO> message) {
        NextTaskFilter filter = task.getNextFilter();

        Map<String, Message> resultMap = new HashMap<>();
        result.stream()
                .map(item -> {
                    HttpRes httpRes = HttpRes.build()
                            .status(item.getStatus())
                            .costTimeMillis(item.getCostTimeMillis())
                            .headers(item.getHeaders())
                            .requestParams(item.getRequestParams());
                    httpRes.setResourceStatus(item.getResourceStatus());
                    httpRes.setTemporary(JsonPathUtil.read(item.getBody(), task.getNextFilter().getFilterArgs()));
                    return httpRes;
                })
                .flatMap(item -> {
                    if (item.getTemporary() instanceof Collection) {
                        return ((Collection<?>) item.getTemporary())
                                .stream()
                                .map(temporary -> {
                                    HttpRes params = HttpRes.build()
                                            .status(item.getStatus())
                                            .costTimeMillis(item.getCostTimeMillis())
                                            .headers(item.getHeaders())
                                            .requestParams(item.getRequestParams());
                                    params.setResourceStatus(item.getResourceStatus());
                                    params.setTemporary(temporary);
                                    return params;
                                });
                    } else {
                        return Stream.of(item);
                    }
                })
                .forEach(httpRes -> {
                            Object temporary = httpRes.getTemporary();
                            httpRes.setTemporary(null);
                            Map map = (Map) temporary;
                            // 是否更新判断
                            String uniqueKey = map.get(filter.getUniqueKey()).toString();
                            uniqueKey = httpRes.getRequestParams().getUrl() + "?uniqueKey=" + uniqueKey;
                            String md5Digest = DigestUtils.md5DigestAsHex(gson.toJson(map).getBytes(StandardCharsets.UTF_8));
                            ResourceStatus resourceStatus = urlMD5Service.getResourceStatus(uniqueKey, md5Digest);
                            urlMD5Service.put(uniqueKey, md5Digest);
                            if (ResourceStatus.UNCHANGED.equals(resourceStatus)) {
                                return;
                            }

                            // 节点命中
                            List<NextTaskFilter.NextNodeConfig> nextNodes = filter.getNextNodes();
                            if (ObjectUtils.isEmpty(nextNodes)) {
                                return;
                            }
                            StandardEvaluationContext context = new StandardEvaluationContext(this);
                            context.setVariable("filterArgs", map);
                            context.setVariable("httpRes", httpRes);

                            for (NextTaskFilter.NextNodeConfig nextNode : nextNodes) {

                                if (!Boolean.TRUE.equals(getPlaceholderSpel(nextNode.getCondition(), filter.getPlaceholderExpressions(), context, Boolean.class))) {
                                    continue;
                                }
                                Set<String> otherCacheKeySet = new HashSet<>();
                                otherCacheKeySet.add(uniqueKey);
                                // 命中
                                String topic = getPlaceholderSpel(nextNode.getTopic(), filter.getPlaceholderExpressions(), context, String.class);
                                Object payload = getPlaceholderSpel(nextNode.getPayload(), filter.getPlaceholderExpressions(), context, Object.class);
                                Message<Object> nextTaskMessage = MessageBuilder
                                        .withPayload(payload)
                                        .setHeader(MessageConstants.CURRENT_STAGE, StageConstants.SOURCE)
                                        .setHeader(MessageConstants.TASK_TRACE_ID, message.getHeaders().get(MessageConstants.TASK_TRACE_ID))
                                        .setHeader(MessageConstants.SOURCE_TRACE_ID, ID.generate())
                                        .setHeader(MessageConstants.TASK_ID, task.getId())
                                        .setHeader(MessageConstants.TASK_NAME, task.getName())
                                        .setHeader(MessageConstants.SEED_URL, httpRes.getRequestParams().getUrl())
                                        .setHeader(MessageConstants.SEED_URL, httpRes.getRequestParams().getUrl())
                                        .setHeader(MessageConstants.OTHER_CACHE_KEY_SET, otherCacheKeySet)
                                        .build();
                                resultMap.put(topic, nextTaskMessage);
                                if ("break".equals(filter.getMode())) {
                                    break;
                                }
                            }
                        }
                );

        return resultMap;

    }

    private void setResourceStatus(HttpRes res) {
        String md5Digest = DigestUtils.md5DigestAsHex(res.getBody().getBytes());
        ResourceStatus resourceStatus = urlMD5Service.getResourceStatus(res.getRequestParams().getUrl(), md5Digest);
        urlMD5Service.put(res.getRequestParams().getUrl(), md5Digest);
        res.setResourceStatus(resourceStatus);
    }


    private List<HttpRes> handleRes(Task task, List<HttpRes> resList) {
        if (ObjectUtils.isEmpty(resList)) {
            return null;
        }
        List<HttpRes> out = new ArrayList<>();
        for (HttpRes res : resList) {
            if (!res.getStatus().is2xxSuccessful()) {
                log.warn("request failed, url: {}, status: {}", res.getRequestParams().getUrl(), res.getStatus());
                continue;
            }
            this.setResourceStatus(res);
            this.log(res, task);
            if (res.getResourceStatus() != ResourceStatus.UNCHANGED || ProcessMode.DEBUG == task.getProcessMode()) {
                out.add(res);
            }
        }
        return out;

    }

    private List<HttpRes> handleRes(Task task, HttpRes res) {
        List<HttpRes> out = new ArrayList<>();
        if (!res.getStatus().is2xxSuccessful()) {
            log.info("log for debug：{}", res);
            log.warn("request failed, url: {}, status: {}", res.getRequestParams().getUrl(), res.getStatus());
            return null;
        }
        this.setResourceStatus(res);
        this.log(res, task);
        if (res.getResourceStatus().equals(ResourceStatus.UNCHANGED) && !ProcessMode.DEBUG.equals(task.getProcessMode())) {
            return null;
        } else {
            out.add(res);
            return out;
        }
    }


    private void log(HttpRes res, Task task) {
        ProcessMode processMode = task.getProcessMode();
        HttpRequestParams req = res.getRequestParams();
        String emoji = "";
        if (processMode.equals(ProcessMode.NORMAL)) {
            emoji = "";
        } else if (processMode.equals(ProcessMode.DEBUG)) {
            emoji = "🤖";
        } else if (processMode.equals(ProcessMode.WATCH)) {
            emoji = "👀";
        }
        if (res.getStatus().is2xxSuccessful()) {
            log.info("[{}-{}] [{}{} {}] [{} {} {}ms] {} ", task.getId(), task.getName(),
                    res.getResourceStatus().getEmoji(), emoji,
                    res.getResourceStatus().getName(), req.getMethod(), res.getStatus(), res.getCostTimeMillis(), req.getUrl());
        } else {
            log.warn("@@ method: [{}], status: [{}], url: [{}], request: [{}], response: [{}]",
                    req.getMethod(), res.getStatus(), req.getUrl(), req.getBody(), res.getBody());
        }
    }

    public HttpRes request(HttpRequestParams req) {
        this.preHandleUrl(req);
        this.preHandleBody(req);
        this.preHandleHeader(req);
        this.preHandleSignature(req);
        return requestClient.execute(req);
    }


    private void preHandleSignature(HttpRequestParams requestParams) {
        HttpRequestParams.RequestSignatureConfig signatureConfig = requestParams.getSignature();
        if (ObjectUtils.isEmpty(signatureConfig)) {
            return;
        }

        String signatureContext = requestParams.getPlaceholderValue(signatureConfig.getSignatureContext(), String.class, parser, new StandardEvaluationContext(this));
        AbstractSignatureService signatureService = null;
        try {
            signatureService = SignatureFactory.getSignatureService(signatureConfig);
            String signature = signatureService.signatureByPriKey(signatureContext);
            // 优化
            if (SignatureEnumGroup.SignatureLocation.header.equals(signatureConfig.getSignatureLocation())) {
                requestParams.getHeaders().put(signatureConfig.getSignatureFile(), signature);
            }
        } catch (GeneralSecurityException e) {
            throw new RuntimeException(e);
        }
    }

    private void preHandleHeader(HttpRequestParams request) {
        Map<String, Object> headers = request.getHeaders();
        if (ObjectUtils.isEmpty(headers)) {
            return;
        }
        request.setHeaders(aclculationPlaceholderMap(request.getHeaders(), request));
    }


    /**
     * 根据占位符从spelMap中获取对应的spel表达式并计算结果
     * 或直接返回占位符（不满足占位符规则时，即无需计算）
     */
    private <T> T getPlaceholderSpel(Object placeholder, Map<String, String> spelMap, EvaluationContext context, Class<T> resultClazz) {
        Matcher matcher = RegexConstants.PLACEHOLDER_PATTERN.matcher(placeholder.toString());
        if (matcher.find()) {
            return parser.parseExpression(spelMap.get(matcher.group())).getValue(context, resultClazz);
        }
        return (T) placeholder;
    }

    /**
     * 计算占位符的值，并返回此map
     */
    private Map<String, Object> aclculationPlaceholderMap(Map<String, Object> placeholderMap, HttpRequestParams request) {

        for (String key : placeholderMap.keySet()) {
            Object placeholderValue = placeholderMap.get(key);
            Matcher matcher = RegexConstants.PLACEHOLDER_PATTERN.matcher(placeholderValue.toString());
            if (matcher.find()) {
                String placeholder = matcher.group();
                Object value = request.getPlaceholderValue(placeholder, Object.class, parser, new StandardEvaluationContext(this));
                placeholderMap.put(key, value);
            }
        }
        return placeholderMap;
    }

    private void preHandleBody(HttpRequestParams request) {
        Map<String, Object> body = request.getBody();
        if (ObjectUtils.isEmpty(body)) {
            return;
        }
        request.setBody(aclculationPlaceholderMap(request.getBody(), request));
//        for (String key : body.keySet()) {
//            Object value = body.get(key);
//            Matcher matcher = RegexConstants.PLACEHOLDER_PATTERN.matcher(value.toString());
//            if (matcher.find()) {
//                String placeholder = matcher.group();
//                String expr = request.getPlaceholderExpressions().get(placeholder);
//                Object result = parser.parseExpression(expr).getValue();
//                body.put(key, result);
//            }
//        }
//        request.setBody(body);
    }

    private void preHandleUrl(HttpRequestParams request) {
        String url = request.getUrl();
        Matcher matcher = RegexConstants.PLACEHOLDER_PATTERN.matcher(url);
        Map<String, Object> patternMap = new HashMap<>();
        while (matcher.find()) {
            String placeholder = matcher.group();
            patternMap.put(placeholder, placeholder);
        }

        if (ObjectUtils.isEmpty(patternMap)) {
            return;
        }


        aclculationPlaceholderMap(patternMap, request);
//        Map<String, String> placeholderExpressions = request.getPlaceholderExpressions();


//        for (String key : patternMap.keySet()) {
//            String expr = placeholderExpressions.get(key);
//            Object value = parser.parseExpression(expr).getValue();
//            patternMap.put(key, value);
//        }
        for (String key : patternMap.keySet()) {
            url = url.replace(key, patternMap.get(key).toString());
        }
        request.setUrl(url);
    }


}



